""" This code is available under the [CC BY 4.0](https://creativecommons.org/licenses/by/4.0/) license.
You must give appropriate credit, provide a link to the license, and indicate if changes were made. You may do so in any
reasonable manner, but not in any way that suggests the licensor endorses you or your use.

Original code by [Sander Vervoort](sander.vervoort@haw-hamburg.de)
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy import signal as scipy_signal


def mzm_transmission(phi: np.array) -> np.array:
    """Mach-Zehnder transmission function

    Args:
        phi (np.array): phase difference between the two arms

    Returns:
        np.array: transmission of the Mach-Zehnder modulator
    """
    return np.cos(phi / 2) ** 2


def triangle(t: np.array, T: float = 2 * np.pi) -> np.array:
    """Triangle function

    Args:
        t (np.array): time
        T (float, optional): period. Defaults to 2*np.pi.

    Returns:
        np.array: _description_
    """
    t = t * np.pi / T
    # pi instead of 2pi because of the abs sawtooth function with width parameter width=0.5
    return np.abs(scipy_signal.sawtooth(t, width=0.5))


def sinc(x: np.array, normalize: bool = False) -> np.array:
    """Sinc function

    Args:
        x (np.array): input
        normalize (bool, optional): if True, normalize by pi sinc = sin(x*pi)/(x*pi)
               else, unnormalized sinc = sin(x)/(x). Defaults to False.

    Returns:
        np.array: sinc(x)
    """
    if not normalize:
        x /= np.pi

    return np.sinc(x)


def calc_sDC(phi_mod: float = np.pi, phi_zero: float = 0.0) -> float:
    """Calculate the DC component of the spectrum

    Args:
        phi_mod (float, optional): modulation phase. Defaults to np.pi.
        phi_zero (float, optional): zero phase. Defaults to 0.0.

    Returns:
        float: DC component of the spectrum
    """
    return (np.sin(phi_mod + phi_zero) - np.sin(phi_zero)) / phi_mod


def calc_Sf(
    f: float, fm: float = 1, phi_mod: float = np.pi, phi_zero: float = 0.0
) -> float:
    """Calculate the spectrum

    Args:
        f (float): frequency
        fm (float, optional): modulation frequency. Defaults to 1.
        phi_mod (float, optional): modulation phase. Defaults to np.pi.
        phi_zero (float, optional): zero phase. Defaults to 0.0.

    Returns:
        float: Signal in frequency domain S(f)
    """
    T = 1 / fm

    return np.abs(
        T
        / 2
        * (
            sinc((np.pi * f * T - phi_mod) / 2)
            * np.cos(phi_zero - (np.pi * f * T - phi_mod) / 2)
            + sinc((np.pi * f * T + phi_mod) / 2)
            * np.cos(phi_zero + (np.pi * f * T + phi_mod) / 2)
        )
        + calc_sDC(phi_mod, phi_zero) * T * sinc(np.pi * f * T)
    )


def calc_sf_sep(
    f: float, fm: float = 1, phi_mod: float = np.pi, phi_zero: float = 0.0
) -> tuple[float, float, float]:
    """Calculate seperate terms of the spectrum

    Args:
        f (float): frequency
        fm (float, optional): modulation frequency. Defaults to 1.
        phi_mod (float, optional): modulation phase. Defaults to np.pi.
        phi_zero (float, optional): zero phase. Defaults to 0.0.

    Returns:
        Tuple[float, float, float]: a, b, and c terms of the spectrum
    """
    T = 1 / fm
    a = (T / 2) * (
        sinc((np.pi * f * T - phi_mod) / 2)
        * np.cos(phi_zero - (np.pi * f * T - phi_mod) / 2)
    )
    b = (T / 2) * (
        sinc((np.pi * f * T + phi_mod) / 2)
        * np.cos(phi_zero + (np.pi * f * T + phi_mod) / 2)
    )
    c = calc_sDC(phi_mod, phi_zero) * T * sinc(np.pi * f * T)
    return a, b, c


def main():
    pdf_export = False
    if pdf_export:
        plt.rcParams.update(
            {
                "text.usetex": True,
                "font.family": "Palatino",
                "figure.figsize": [10 / 2.54, 6 / 2.54],  # 10cm x 6cm
                "savefig.bbox": "tight",  # prevent missing labels
            }
        )

    phi_mod = 1.2 * np.pi
    phi_zero = 0.8 * np.pi
    x = np.linspace(-1.5, 1.5, 301, endpoint=True)  # t/T

    ## time domain phase plot
    phi = (1 - triangle(x, T=1)) * phi_mod + phi_zero  # flip triangle
    labels = [r"$\phi_0$", r"$\phi_0 + \phi_\mathrm{m}$"]

    plt.plot(x, phi, "--C0")
    plt.plot(x[100:200], phi[100:200], "-C0")

    plt.xlim([-1.25, 1.25])
    plt.xlabel(r"$t\,/\,T_\mathrm{m}$")
    plt.yticks([phi_zero, phi_mod + phi_zero], labels)
    plt.ylabel(r"$\phi(t)$")
    plt.grid(True)

    if pdf_export:
        plt.savefig("figures/mzm_time_domain_phase_plot.pdf")
    plt.show()

    ## time domain signal plot
    s = mzm_transmission(phi)

    plt.plot(x, s, "--C0")
    plt.plot(x[100:200], s[100:200], "-C0")

    plt.xlim([-1.25, 1.25])
    plt.xlabel(r"$t\,/\,T_\mathrm{m}$")
    plt.ylabel(r"$\tau_\mathrm{a}(t)$")
    plt.yticks([0, 1])
    plt.grid(True)

    if pdf_export:
        plt.savefig("figures/mzm_time_domain_plot.pdf")
    plt.show()

    ## analytical spectrum

    # continuous spectrum
    f = np.linspace(0, 5, 200 + 1)
    S = calc_Sf(f, phi_mod=phi_mod, phi_zero=phi_zero)
    plt.plot(f, S, "C2")
    # discrete spectrum
    f = np.linspace(0, 5, 5 + 1)
    S = calc_Sf(f, phi_mod=phi_mod, phi_zero=phi_zero)
    plt.stem(f, S)

    plt.xlabel(r"$f\,/\,f_{m}$")
    plt.ylabel(r"Norm.\ Amplitude")

    plt.legend(
        [
            r"$\left|\underline{S}(f)\right| / T_\mathrm{m}$",
            r"$\left|\underline{S}_\mathrm{m}(f)\right|$",
        ]
    )

    if pdf_export:
        plt.savefig("figures/mzm_analystical_spectrum.pdf")
    plt.show()

    ## analytical spectrum with separated terms
    if pdf_export:
        plt.rcParams.update({"figure.figsize": [15 / 2.54, 9 / 2.54]})  # 16cm x 8cm

    f = np.linspace(-5, 5, 400 + 1)
    Sa, Sb, Sc = calc_sf_sep(f, phi_mod=phi_mod, phi_zero=phi_zero)
    plt.plot(f, Sa, "--C2")
    plt.plot(f, Sb, "-.C2")
    plt.plot(f, Sc, ":C2")

    # continuous spectrum
    S = calc_Sf(f, phi_mod=phi_mod, phi_zero=phi_zero)
    plt.plot(f, S, "C2")
    # discrete spectrum
    f = np.linspace(-5, 5, 10 + 1)
    S = calc_Sf(f, phi_mod=phi_mod, phi_zero=phi_zero)
    plt.stem(f, S)

    plt.xlim([-4.8, 4.8])
    plt.xlabel(r"$f\,/\,f_{m}$")
    plt.ylabel(r"Norm.\ Amplitude")

    plt.legend(
        [
            "$C / T_\mathrm{m}$",
            "$B / T_\mathrm{m}$",
            "$A / T_\mathrm{m}$",
            r"$\left|\underline{S}(f)\right| / T_\mathrm{m}$",
            r"$\left|\underline{S}_\mathrm{p}(f)\right|$",
        ],
        loc="upper left",
    )
    plt.grid(True)

    if pdf_export:
        plt.savefig("figures/mzm_analystical_spectrum_parts.pdf")
    plt.show()


if __name__ == "__main__":
    main()
